//	GamePlay.cpp - Main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2011  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "GamePlay.h"
#include "FilePath.h"
#include "NiallsDrawingLib/OpenGLDrawer.h"
#include "NiallsDrawingLib/ImageLoader.h"
#include "Images.h"
#include "Random.h"
#include "ImageManager.h"
#include "PostProc.h"
#include "Sounds.h"

#include <cmath>
#include <sstream>

using namespace std;

//------------------------------------------------------------------------------
uint32_t GamePlay::titleImage = 0;
uint32_t GamePlay::ritualsImage = 0;
uint32_t GamePlay::divineBackImage = 0;
uint32_t GamePlay::divineSunImage = 0;
uint32_t GamePlay::divineLinesImage = 0;

//------------------------------------------------------------------------------
GamePlay::GamePlay(Application *app):
GameState(app),
titleText(TwoFloats(24.0f, 32.0f)),
ritualsText(TwoFloats(1280.0f-900.0f, 720.0f-128.0f)),
ghostFade(0.0f),
indicesCount(30),
currentNoise(0),
noiseCount(3),
sunPulse(0.0f),
ritualsDelay(120),
nextKey(SDLK_0),
nextNumber(0),
numShantis(0),
shantiIndex(0),
divineFade(0.0f),
firstRun(true),
funTimes(false)
{
	int x, y;
	float gridX, gridY;

	ritualVals[0] = 0.0f;
	ritualVals[1] = 0.0f;
	ritualVals[2] = 0.0f;

	ritualSpeeds[0] = -0.0005f;
	ritualSpeeds[1] = -0.0005f;
	ritualSpeeds[2] = -0.0005f;

	ritualStrings[0] = "hack";
	ritualStrings[1] = "scandal";
	ritualStrings[2] = "paedophile";

	ritualIndices[0] = 0;
	ritualIndices[1] = 0;
	ritualIndices[2] = 0;

	gridX = (1280.0f/GridWidth) * 1.25f;
	gridY = (720.0f/GridHeight) * 1.25f;

	for(y=0;y<GridHeight;++y)
	{
		for(x=0;x<GridWidth;++x)
		{
			images[(y*GridWidth)+x].setPosition(TwoFloats((gridX*x)-(gridX*0.5f),
														  (gridY*y)/*-(gridY*0.5f)*/));
		}
	}

	mixUpIndices();
}

//------------------------------------------------------------------------------
GamePlay::~GamePlay()
{
	
}

//------------------------------------------------------------------------------
void GamePlay::draw(Drawer *d)
{
	int i;
	int x, y;
	int tempint;
	float tempLev;
	float sunSize;
	float stringFade;

	if(firstRun)
	{
		funTimes = d->canSupport(Drawer::Shaders) && d->canSupport(Drawer::FrameBuffer);
		firstRun = false;
	}

	if(!divineBackImage)
	{
		ImageData data;
		ImageLoader loader;

		data = loader.loadImage(Images::title_png, Images::title_pngSize);
		titleImage = d->loadImage(data.data, data.width, data.height);

		data = loader.loadImage(Images::rituals_png, Images::rituals_pngSize);
		ritualsImage = d->loadImage(data.data, data.width, data.height);

		data = loader.loadImage(Images::divineback_jpg, Images::divineback_jpgSize);
		divineBackImage = d->loadImage(data.data, data.width, data.height);

		data = loader.loadImage(Images::sun_png, Images::sun_pngSize);
		divineSunImage = d->loadImage(data.data, data.width, data.height);

		data = loader.loadImage(Images::divinelines_png, Images::divinelines_pngSize);
		divineLinesImage = d->loadImage(data.data, data.width, data.height);
	}

	ImageManager::getInstance().loadImages(d);

	//Mix up imageIndices if necessary.
	--indicesCount;
	if(indicesCount < 0)
	{
		indicesCount = 90 + (int)(Random::getGlobal().getRandomFloat() * 29.99f);

		mixUpIndices();
	}

	//Update the circle detector.
	/*if(circleDetector.update())
		ritualVals[1] += 0.0025f;

	//Decay ritual vals if necessary.
	if((ritualVals[1] > 0.0f) && (ritualVals[1] < 0.9f))
		ritualVals[1] -= 0.0005f;*/

	/*if(ritualVals[0] > 1.0f)
		ritualVals[0] = 1.0f;
	if(ritualVals[1] > 1.0f)
		ritualVals[1] = 1.0f;
	if(ritualVals[2] > 1.0f)
		ritualVals[2] = 1.0f;*/

	//Handle the typing rituals.
	//Numbers.
	if(ritualSpeeds[0] > 0.0033f)
		ritualSpeeds[0] = 0.0033f;
	else if(ritualSpeeds[0] > -0.0005f)
		ritualSpeeds[0] -= 0.0001f;

	if(ritualVals[0] < 1.0f)
	{
		ritualVals[0] += ritualSpeeds[0];
		if(ritualVals[0] >= 1.0f)
			nextKey = SDLK_s;
	}
	if(ritualVals[0] < 0.0f)
		ritualVals[0] = 0.0f;

	if(ritualSpeeds[1] > 0.0033f)
		ritualSpeeds[1] = 0.0033f;
	else if(ritualSpeeds[1] > -0.0005f)
		ritualSpeeds[1] -= 0.0001f;

	if(ritualVals[1] < 1.0f)
	{
		ritualVals[1] += ritualSpeeds[1];
		if(ritualVals[1] >= 1.0f)
			nextKey = SDLK_s;
	}
	if(ritualVals[1] < 0.0f)
		ritualVals[1] = 0.0f;

	//Shanti.
	if(ritualSpeeds[2] > 0.0033f)
		ritualSpeeds[2] = 0.0033f;
	else if(ritualSpeeds[2] > -0.0005f)
		ritualSpeeds[2] -= 0.0001f;

	if(ritualVals[2] < 1.0f)
		ritualVals[2] += ritualSpeeds[2];
	if(ritualVals[2] < 0.0f)
		ritualVals[2] = 0.0f;

	if(((ritualVals[0] + ritualVals[1] + ritualVals[2]) > 2.7f) && divineFade < 1.0f)
	{
		divineFade += 0.005f;
		if(divineFade >= 1.0f)
		{
			Sounds::getInstance().playSound(Sounds::FinalSpeech1);
			Sounds::getInstance().playSound(Sounds::FinalSpeech2);
		}
		if(ritualVals[0] < 1.0f)
			ritualVals[0] += 0.005f;
		if(ritualVals[1] < 1.0f)
			ritualVals[1] += 0.005f;
		if(ritualVals[2] < 1.0f)
			ritualVals[2] += 0.005f;
	}
	else if(divineFade >= 1.0f)
	{
		ritualVals[0] = 1.0f;
		ritualVals[1] = 1.0f;
		ritualVals[2] = 1.0f;
	}

	//Update StreamingSound.
	sounds.update(ritualVals[0], ritualVals[1], ritualVals[2]);

	//Apply correct hide values to the GlitchImages.
	//Ritual 2 (shanti).
	for(y=0;y<GridHeight;++y)
	{
		for(x=0;x<GridWidth;++x)
			images[(y*GridWidth)+x].setHide(ritualVals[2]);
	}
	//Ritual 1 (circle).
	for(y=Ritual1StartY;y<Ritual1EndY;++y)
	{
		for(x=Ritual1StartX;x<Ritual1EndX ;++x)
			images[(y*GridWidth)+x].setHide(ritualVals[1]);
	}
	//Ritual 0 (qabbalah).
	for(y=Ritual0StartY;y<Ritual0EndY;++y)
	{
		for(x=Ritual0StartX;x<Ritual0EndX ;++x)
			images[(y*GridWidth)+x].setHide(ritualVals[0]);
	}

	//Calculate the size of the sun.
	sunPulse += 0.033f;
	if(sunPulse > 6.283185f)
		sunPulse -= 6.283185f;
	sunSize = (740.0f + (sinf(sunPulse) * 64.0f))/1024.0f;

	//Update DivineStrings.
	stringFade = ritualVals[0] + ritualVals[1] + ritualVals[2];
	if(stringFade < 2.5f)
		stringFade = 0.0f;
	else
		stringFade = (stringFade - 2.5f) * 2.0f;
	if(stringFade > 0.0f)
		tempLev = (stringFade * 0.5f) + (sinf(sunPulse) * 0.25f);
	else
		tempLev = 0.0f;
	divineStrings.setLevel(tempLev);

	//Set ghostFade.
	if(ritualVals[0] < 1.0f)
		ghostFade = 0.0f;
	else if(stringFade > 0.0f)
		ghostFade = 1.0f-stringFade;
	else if(ghostFade < 1.0f)
		ghostFade += 0.001f;

	if(funTimes)
		d->startFrameBufferDrawing();

	//Draw the divine.
	float tempf1 = 0.1f + (divineFade * 0.9f);
	float tempf2 = 0.05f + (divineFade * 0.95f);
	d->drawImage(divineBackImage, 0, 0, 0.625f, 0.703125f, divineFade, tempf2, tempf1);
	d->drawImage(divineBackImage, 1280, 0, -0.625f, 0.703125f, divineFade, tempf2, tempf1);
	d->drawImage(divineSunImage,
				 640-(sunSize * 0.5f * 1024.0f),
				 360-(sunSize * 0.5f * 1024.0f),
				 sunSize,
				 sunSize,
				 1.0f,
				 1.0f,
				 1.0f,
				 divineFade);
	d->drawImage(divineLinesImage, 0, 0, 1280.0f);

	//Draw the ghosts.
	if(ghostFade > 0.0f)
	{
		for(i=0;i<NumGhosts;++i)
			ghosts[i].draw(d, ghostFade, funTimes);
	}

	//Draw the glitched images.
	for(i=0;i<(GridWidth*GridHeight);++i)
		images[imageIndices[i]].draw(d, funTimes);

	//Draw any text.
	titleText.draw(d, titleImage);
	if(ritualsDelay > 0)
		--ritualsDelay;
	else
		ritualsText.draw(d, ritualsImage);

	//Temp.
	//circleDetector.draw(d);

	//Post-processing.
	if(funTimes)
	{
		d->endFrameBufferDrawing();

		--noiseCount;
		if(noiseCount < 0)
		{
			noiseCount = 5;
			tempint = currentNoise;
			while(tempint == currentNoise)
				tempint = (int)(Random::getGlobal().getRandomFloat() * 3.99f);
			currentNoise = tempint;
		}

		d->useSecondImage(PostProc::images[currentNoise]);
		d->useShader(PostProc::shader);
		d->drawImage(d->getFrameBufferImage(), 0, 0);
		d->useShader(0);
		d->useSecondImage(0);
	}
}

//------------------------------------------------------------------------------
void GamePlay::getAudio(float *buffer, int numSamples)
{
	sounds.getAudio(buffer, numSamples);
	divineStrings.getAudio(buffer, numSamples);
}

//------------------------------------------------------------------------------
void GamePlay::keyDown(const string& action, const SDL_keysym& keyPressed)
{
	if(action == "Escape")
		quit();
	else
	{
		char tempch = getChar(keyPressed.sym);

		if((ritualVals[0] < 1.0f) &&
		   (tempch == ritualStrings[0][ritualIndices[0]]))
		{
			ritualSpeeds[0] += 0.002f;
			++ritualIndices[0];
			if(ritualIndices[0] >= ritualStrings[0].length())
				ritualIndices[0] = 0;
		}
		else if((ritualVals[1] < 1.0f) &&
			   (tempch == ritualStrings[1][ritualIndices[1]]))
		{
			ritualSpeeds[1] += 0.002f;
			++ritualIndices[1];
			if(ritualIndices[1] >= ritualStrings[1].length())
				ritualIndices[1] = 0;
		}
		else if((ritualVals[2] < 1.0f) &&
			   (tempch == ritualStrings[2][ritualIndices[2]]))
		{
			ritualSpeeds[2] += 0.002f;
			++ritualIndices[2];
			if(ritualIndices[2] >= ritualStrings[2].length())
				ritualIndices[2] = 0;
		}
	}
}

//------------------------------------------------------------------------------
void GamePlay::mouseMove(int x, int y)
{
	//circleDetector.setMousePos(TwoFloats(x, y));
}

//------------------------------------------------------------------------------
void GamePlay::mixUpIndices()
{
	int i;
	int tempint;
	vector<int> tempIndices;

	for(i=0;i<(GridWidth*GridHeight);++i)
		tempIndices.push_back(i);

	for(i=0;i<(GridWidth*GridHeight);++i)
	{
		tempint = (int)(Random::getGlobal().getRandomFloat() * (tempIndices.size() * 0.99999f));
		imageIndices[i] = tempIndices[tempint];
		tempIndices.erase(tempIndices.begin() + tempint);
	}
}

//------------------------------------------------------------------------------
bool GamePlay::isNumber(SDLKey key)
{
	bool retval = false;

	switch(key)
	{
		case SDLK_0:
		case SDLK_1:
		case SDLK_2:
		case SDLK_3:
		case SDLK_4:
		case SDLK_5:
		case SDLK_6:
		case SDLK_7:
		case SDLK_8:
		case SDLK_9:
		case SDLK_KP0:
		case SDLK_KP1:
		case SDLK_KP2:
		case SDLK_KP3:
		case SDLK_KP4:
		case SDLK_KP5:
		case SDLK_KP6:
		case SDLK_KP7:
		case SDLK_KP8:
		case SDLK_KP9:
			retval = true;
			break;
		default:
			retval = false;
			break;
	}

	return retval;
}

//------------------------------------------------------------------------------
bool GamePlay::isShanti(SDLKey key)
{
	bool retval = false;

	switch(key)
	{
		case SDLK_s:
		case SDLK_h:
		case SDLK_a:
		case SDLK_n:
		case SDLK_t:
		case SDLK_i:
			retval = true;
			break;
		default:
			retval = false;
			break;
	}

	return retval;
}

//------------------------------------------------------------------------------
void GamePlay::incrementNextKeyNumber()
{
	if((nextKey == SDLK_1) && (nextNumber == 10))
		nextKey = SDLK_0;
	else if((nextKey == SDLK_0) && (nextNumber == 10))
	{
		nextKey = SDLK_0;
		nextNumber = 0;
	}
	else
	{
		switch(nextKey)
		{
			case SDLK_0:
				nextKey = SDLK_1;
				break;
			case SDLK_1:
				nextKey = SDLK_2;
				break;
			case SDLK_2:
				nextKey = SDLK_3;
				break;
			case SDLK_3:
				nextKey = SDLK_4;
				break;
			case SDLK_4:
				nextKey = SDLK_5;
				break;
			case SDLK_5:
				nextKey = SDLK_6;
				break;
			case SDLK_6:
				nextKey = SDLK_7;
				break;
			case SDLK_7:
				nextKey = SDLK_8;
				break;
			case SDLK_8:
				nextKey = SDLK_9;
				break;
			case SDLK_9:
				nextKey = SDLK_1;
				break;
		}
		++nextNumber;
	}

	if(ritualVals[0] < 1.0f)
		ritualSpeeds[0] += 0.002f;

	/*if(ritualVals[0] >= 1.0f)
		nextKey = SDLK_s;*/
}

//------------------------------------------------------------------------------
void GamePlay::incrementNextKeyLetter()
{
	switch(nextKey)
	{
		case SDLK_s:
			nextKey = SDLK_h;
			break;
		case SDLK_h:
			if((numShantis >= 2) && (shantiIndex > 6))
				nextKey = SDLK_i;
			else
				nextKey = SDLK_a;
			break;
		case SDLK_a:
			nextKey = SDLK_n;
			break;
		case SDLK_n:
			nextKey = SDLK_t;
			break;
		case SDLK_t:
			nextKey = SDLK_i;
			break;
		case SDLK_i:
			if(numShantis < 2)
			{
				nextKey = SDLK_s;
				++numShantis;
				shantiIndex = -1;
			}
			else if(shantiIndex == 5)
				nextKey = SDLK_h;
			else if(shantiIndex > 5)
			{
				nextKey = SDLK_s;
				++numShantis;
				shantiIndex = -1;
			}
			break;
	}

	++shantiIndex;

	if(ritualVals[2] < 1.0f)
		ritualSpeeds[2] += 0.002f;
}

//------------------------------------------------------------------------------
char GamePlay::getChar(SDLKey sym)
{
	char retval;

	switch(sym)
	{
		case SDLK_a:
			retval = 'a';
			break;
		case SDLK_b:
			retval = 'b';
			break;
		case SDLK_c:
			retval = 'c';
			break;
		case SDLK_d:
			retval = 'd';
			break;
		case SDLK_e:
			retval = 'e';
			break;
		case SDLK_f:
			retval = 'f';
			break;
		case SDLK_g:
			retval = 'g';
			break;
		case SDLK_h:
			retval = 'h';
			break;
		case SDLK_i:
			retval = 'i';
			break;
		case SDLK_j:
			retval = 'j';
			break;
		case SDLK_k:
			retval = 'k';
			break;
		case SDLK_l:
			retval = 'l';
			break;
		case SDLK_m:
			retval = 'm';
			break;
		case SDLK_n:
			retval = 'n';
			break;
		case SDLK_o:
			retval = 'o';
			break;
		case SDLK_p:
			retval = 'p';
			break;
		case SDLK_q:
			retval = 'q';
			break;
		case SDLK_r:
			retval = 'r';
			break;
		case SDLK_s:
			retval = 's';
			break;
		case SDLK_t:
			retval = 't';
			break;
		case SDLK_u:
			retval = 'u';
			break;
		case SDLK_v:
			retval = 'v';
			break;
		case SDLK_w:
			retval = 'w';
			break;
		case SDLK_x:
			retval = 'x';
			break;
		case SDLK_y:
			retval = 'y';
			break;
		case SDLK_z:
			retval = 'z';
			break;
		default:
			retval = ' ';
	}

	return retval;
}
