//	GamePlay.h - Main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2011  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GAMEPLAY_H_
#define GAMEPLAY_H_

#include "GameState.h"
#include "GhostImage.h"
#include "GlitchText.h"
#include "GlitchImage.h"
#include "DivineStrings.h"
#include "CircleDetector.h"
#include "StreamingSound.h"

#include <stdint.h>

///	Main gameplay state.
class GamePlay : public GameState
{
  public:
	///	Constructor.
	GamePlay(Application *app);
	///	Destructor.
	~GamePlay();

	///	Draws the state.
	void draw(Drawer *d);

	///	Generates any state-specific audio.
	void getAudio(float *buffer, int numSamples);

	///	Called when a key down event is received.
	void keyDown(const std::string& action, const SDL_keysym& keyPressed);

	///	Called when a mouse move event is received.
	void mouseMove(int x, int y);

	///	Returns the name of this state.
	const std::string getName() const {return "GamePlay";};
  private:
	///	Helper method. Switches up imageIndices.
	void mixUpIndices();
	///	Helper method. Returns true if key is a number key.
	bool isNumber(SDLKey key);
	///	Helper method. Returns true if key is one of the shantihi letters.
	bool isShanti(SDLKey key);
	///	Helper method. Updates nextKey.
	void incrementNextKeyNumber();
	///	Helper method. Updates nextKey.
	void incrementNextKeyLetter();
	///	Helper method. Returns the char corresponding to an SDL keysym.
	char getChar(SDLKey sym);

	///	Title text image.
	static uint32_t titleImage;
	///	Rituals text image.
	static uint32_t ritualsImage;
	///	The divine image.
	static uint32_t divineBackImage;
	///	The divine image.
	static uint32_t divineSunImage;
	///	The divine image.
	static uint32_t divineLinesImage;

	///	Dimensions of the image grid.
	enum
	{
		GridWidth = 12,
		GridHeight = 8,

		NumGhosts = 8,

		Ritual0StartX = 4,
		Ritual0StartY = 2,
		Ritual0EndX = 8,
		Ritual0EndY = 6,

		Ritual1StartX = 2,
		Ritual1StartY = 1,
		Ritual1EndX = 10,
		Ritual1EndY = 7
	};

	///	The title text.
	GlitchText titleText;
	///	The rituals text.
	GlitchText ritualsText;
	///	The glitched images.
	GlitchImage images[GridHeight*GridWidth];
	///	The order to draw the glitched images in.
	int imageIndices[GridHeight*GridWidth];

	///	The GhostImages we display once the 1st ritual's gone.
	GhostImage ghosts[NumGhosts];
	///	Used to fade in the ghosts.
	float ghostFade;

	///	Used to switch up imageIndices.
	int indicesCount;

	///	Used to switch between noise images.
	int currentNoise;
	///	Used to switch between noise images.
	int noiseCount;

	///	Used to make the divine sun pulsate.
	float sunPulse;

	///	Used to delay displaying ritualsText for a wee while.
	int ritualsDelay;

	///	The 3 ritual values.
	float ritualVals[3];
	///	The 3 speeds of ritualVals updates.
	float ritualSpeeds[3];
	///	Used to handle the counting ritual.
	SDLKey nextKey;
	///	Which number we're waiting for.
	int nextNumber;
	///	How many shantis the player has typed so far.
	int numShantis;
	///	Where we are in the current shanti.
	int shantiIndex;
	///	Used to detect when the player's drawing circles with the mouse.
	CircleDetector circleDetector;

	///	The three ritual texts.
	std::string ritualStrings[3];
	///	The indices for the three ritual texts.
	int ritualIndices[3];

	///	Used to fade in the divine.
	float divineFade;

	///	The main collection of sounds.
	StreamingSound sounds;
	///	The string models for The Divine.
	DivineStrings divineStrings;

	///	Used to ensure we only call Drawer::canSupport() once.
	bool firstRun;
	///	Whether or not the user's graphics card can support shaders and the framebuffer.
	bool funTimes;
};

#endif
