//	GlitchImage.cpp - A glitching image class.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2011  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "GlitchImage.h"
#include "NiallsDrawingLib/Drawer.h"
#include "NiallsDrawingLib/ImageLoader.h"
#include "ImageManager.h"
#include "Random.h"
#include "FilePath.h"
#include "PostProc.h"

#include <cassert>

using namespace std;

//------------------------------------------------------------------------------
int GlitchImage::loadingStagger = 0;
uint32_t GlitchImage::featherShader = 0;

//------------------------------------------------------------------------------
GlitchImage::GlitchImage():
image(0),
hide(0.0f)
{
	int i;
	float y = 0;

	++loadingStagger;
	newImageCount = loadingStagger;// + (int)(Random::getGlobal().getRandomFloat() * 29.999f);
	sectionCount = 20 + (int)(Random::getGlobal().getRandomFloat() * 19.999f);

	for(i=0;i<NumSections;++i)
	{
		sections[i].x = (Random::getGlobal().getRandomFloat() * 192.0f) - 96.0f;

		sections[i].y = y;
		sectionHeights[i] = (Random::getGlobal().getRandomFloat() * (512.0f/(float)NumSections));
		y += sectionHeights[i];
	}
}

//------------------------------------------------------------------------------
GlitchImage::~GlitchImage()
{

}

//------------------------------------------------------------------------------
void GlitchImage::draw(Drawer *d, bool funTimes)
{
	int i;
	TwoFloats tempPos = position;

	if(hide >= 1.0f)
		return;

	if(!PostProc::featherImages.size())
	{
		FilePath shaderPath(FilePath::getDataFilesDir().getChild("shaders"));
		FilePath feathersPath = FilePath::getDataFilesDir().getChild("feathers");

		assert(feathersPath.getNumChildren() > 0);

		//Load the feather images.
		for(i=0;i<feathersPath.getNumChildren();++i)
		{
			ImageData data;
			ImageLoader loader;

			data = loader.loadImage(feathersPath.getChild(i).getPath());
			PostProc::featherImages.push_back(d->loadImage(data.data, data.width, data.height));
		}

		//Load the feather shader.
		if(funTimes)
		{
			featherShader = d->loadShader(shaderPath.getChild("feather.vert").getPath(),
										  shaderPath.getChild("feather.frag").getPath());
		}
	}

	if(!image)
	{
		image = ImageManager::getInstance().getRandomImage();
		featherIndex = (int)(Random::getGlobal().getRandomFloat() * ((float)PostProc::featherImages.size()-0.001f));
	}

	--newImageCount;
	if(newImageCount < 0)
	{
		newImageCount = 30 + (int)(Random::getGlobal().getRandomFloat() * 29.999f);

		image = ImageManager::getInstance().getRandomImage();
		featherIndex = (int)(Random::getGlobal().getRandomFloat() * ((float)PostProc::featherImages.size()-0.001f));
	}

	--sectionCount;
	if(sectionCount < 0)
	{
		sectionCount = 20 + (int)(Random::getGlobal().getRandomFloat() * 19.999f);

		randomiseSections();
	}

	if(hide > 0.0f)
	{
		TwoFloats normal;
		const TwoFloats centre(640.0f, 360.0f);

		normal = (position - centre).normal();

		tempPos = position + (hide * normal * centre);
	}

	if(funTimes)
	{
		d->useSecondImage(PostProc::featherImages[featherIndex]);
		d->useShader(featherShader);
	}

	for(i=0;i<NumSections;++i)
	{
		d->drawSubImage(image,
						tempPos.x - 256 + sections[i].x,
						tempPos.y - 256 + sections[i].y,
						0.0f,
						(sections[i].y/512.0f),
						1.0f,
						(sectionHeights[i]/512.0f),
						1.0f,
						1.0f,
						1.0f,
						1.0f,
						1.0f,
						(1.0f-hide));
	}

	if(funTimes)
	{
		d->useShader(0);
		d->useSecondImage(0);
	}
}

//------------------------------------------------------------------------------
void GlitchImage::setPosition(const TwoFloats& pos)
{
	position = pos;
}

//------------------------------------------------------------------------------
void GlitchImage::setHide(float val)
{
	hide = val;
}

//------------------------------------------------------------------------------
void GlitchImage::randomiseSections()
{
	int i;
	float y = 0;

	if(Random::getGlobal().getRandomFloat() > 0.25f)
	{
		for(i=0;i<NumSections;++i)
		{
			sections[i].x = (Random::getGlobal().getRandomFloat() * 192.0f) - 96.0f;

			sections[i].y = y;
			sectionHeights[i] = (Random::getGlobal().getRandomFloat() * (512.0f/(float)NumSections));
			y += sectionHeights[i];
		}
	}
	else
	{
		for(i=0;i<NumSections;++i)
		{
			sections[i].x = 0.0f;

			sections[i].y = y;
			sectionHeights[i] = (Random::getGlobal().getRandomFloat() * (512.0f/(float)NumSections));
			y += sectionHeights[i];
		}
	}
}
