//	OpenGLFont.cpp - Used to draw truetype fonts.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2011  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "OpenGLFont.h"
#include "DebugStatement.h"

#ifdef USE_TTF
#include "SDL_ttf.h"
#endif

#include <iostream>
#include <cmath>

using namespace std;

//------------------------------------------------------------------------------
OpenGLFont::OpenGLFont(const string& name, float maxSize):
atlas(0),
valid(false)
{
#ifdef USE_TTF
	int i;
	float y;
	float tempf;
	int lineFonts;
	TTF_Font *font;
	int numFontsPerLine;
	float maxWidth, maxHeight;
	int minX, maxX, minY, maxY, advance;
	SDL_Color white = {255, 255, 255};
	FilePath phil = getFontPath(name);

	{
		GLenum err = glGetError();
		DebugStatement("OpenGLFont: Initial error status = " << gluErrorString(err));
	}

	if(phil.exists())
	{
		//Initialise SDL_TTF.
		//TTFSingleton::getInstance();

		//Load the font.
		font = TTF_OpenFont(phil.getPath().c_str(), (int)maxSize);

		if(font)
		{
			maxWidth = 0.0f;
			maxHeight = 0.0f;

			//Find the maximum width and height of a glyph.
			for(i=0;i<charactersSize;++i)
			{
				TTF_GlyphMetrics(font,
								 characters[i],
								 &minX,
								 &maxX,
								 &minY,
								 &maxY,
								 &advance);
				if((float)advance > maxWidth)
					maxWidth = (float)advance;
				if((float)(maxY-minY) > maxHeight)
					maxHeight = (float)(maxY-minY);
			}
			int skip = TTF_FontLineSkip(font);
			int h = TTF_FontHeight(font);
			height = (skip > h) ? skip : h;
			height = (height > maxHeight) ? height : maxHeight;
			height += 4;

			//Calculate the size of the texture atlas.
			tempf = sqrtf((float)charactersSize);
			numFontsPerLine = (int)tempf+1;
			atlasSize = (numFontsPerLine*((maxWidth > height) ? maxWidth : height));
			atlasSize = powerOfTwo(atlasSize);

			DebugStatement("font " << name << ":" << maxSize << " atlasSize = " << atlasSize);
			if(atlasSize > 1024)
				DebugStatement("OpenGLFont: Warning: Font atlas dimensions greater than 1024: (" << atlasSize << ")");

#if __BIG_ENDIAN__==1
		DebugStatement("OpenGLFont: Big endian architecture, flipping bytes.");
#else
		DebugStatement("OpenGLFont: Little endian architecture, not flipping bytes.");
#endif

			//Create the texture atlas image data.
			const uint32_t dataSize = atlasSize*atlasSize*4;
			uint8_t *atlasData = new uint8_t[dataSize];

			//Blank atlasData.
			for(i=0;i<dataSize;i+=4)
			{
				atlasData[i] = 255;
				atlasData[i+1] = 255;
				atlasData[i+2] = 255;
				atlasData[i+3] = 0;
			}

			//Copy the glyphs to atlasData.
			y = 0;
			lineFonts = 0;
			for(i=0;i<charactersSize;++i)
			{
				Glyph g;
				SDL_Surface *glyph;

				//Get the glyph's width, and yOffset.
				TTF_GlyphMetrics(font,
								 characters[i],
								 &minX,
								 &maxX,
								 &minY,
								 &maxY,
								 &advance);

				//Render the glyph.
				glyph = TTF_RenderGlyph_Blended(font, characters[i], white);

				//Copy it to atlasData.
				copyGlyph(atlasData, glyph, lineFonts*maxWidth, y, maxHeight-maxY);

				//Add the glyph's data to glyphs.
				g.character = characters[i];
				g.xPos = (float)(lineFonts*maxWidth);
				g.yPos = y;
				if(advance >= (maxX-minX))
				{
					g.width = advance;
					g.xOffset = 0.0f;
					g.advance = advance;
				}
				else
				{
					g.width = (maxX-minX);
					g.xOffset = -(g.width-advance)*0.5f;
					g.advance = advance;
				}
				glyphs.push_back(g);

				//Free the glyph.
				SDL_FreeSurface(glyph);

				++lineFonts;
				if(lineFonts > numFontsPerLine)
				{
					y += height+4;
					lineFonts = 0;
				}
			}

			//Copy atlasData onto the graphics card.
			glGenTextures(1, &atlas);

			glBindTexture(GL_TEXTURE_2D, atlas);

			glTexImage2D(GL_TEXTURE_2D,
						 0,
						 GL_RGBA,
						 atlasSize,
						 atlasSize,
						 0,
						 GL_RGBA,
						 GL_UNSIGNED_BYTE,
						 atlasData);
#ifdef DEBUG_STATEMENTS
			GLenum err = glGetError();
			if(err != GL_NO_ERROR)
				DebugStatement("OpenGLFont: GL error = " << gluErrorString(err));
#endif

			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

			glBindTexture(GL_TEXTURE_2D, 0);

			//Get rid of atlasData.
			delete [] atlasData;

			//Get rid of the font.
			TTF_CloseFont(font);

			valid = true;
		}
		else
			cout << "OpenGLFont: Could not load font: " << name << endl;
	}
	else
		cout << "OpenGLFont: Could not find font: " << name << endl;
#endif
}

//------------------------------------------------------------------------------
OpenGLFont::~OpenGLFont()
{
	
}

//------------------------------------------------------------------------------
void OpenGLFont::draw(const string& text,
					  float x,
					  float y,
					  float boxWidth,
					  float scale,
					  float r,
					  float g,
					  float b,
					  float a)
{
#ifdef USE_TTF
	int index;
	float span;
	float tempX;
	int i, j, k;
	bool foundSpace;
	vector<Glyph *> line;

	span = 0.0f;
	tempX = 0.0f;
	foundSpace = false;
	for(i=0;i<text.length();++i)
	{
		//Add line breaks if necessary.
		if(text[i] == '\n')
		{
			//Draw the line.
			tempX = x;
			for(k=0;k<line.size();++k)
			{
				if(line[k])
				{
					drawGlyph(line[k]->xPos,
							  line[k]->yPos,
							  line[k]->width,
							  scale,
							  tempX + line[k]->xOffset,
							  y,
							  r,
							  g,
							  b,
							  a);
					tempX += (line[k]->advance * scale);
				}
				else
					tempX += (glyphs[getGlyphIndex(' ')].advance * scale);
			}

			y += height;
			span = 0.0f;
			line.clear();
		}
		//Add the glyph to line, increment span accordingly.
		else
		{
			index = getGlyphIndex(text[i]);
			if(index < charactersSize)
			{
				line.push_back(&(glyphs[index]));
				span += (glyphs[index].advance * scale);
			}
			else
			{
				line.push_back(0);
				span += (glyphs[getGlyphIndex(' ')].advance * scale);
			}

			//Once we've reached the other end of the bounding box, draw the line.
			if(span >= boxWidth)
			{
				//Count back to the first space.
				for(j=(line.size()-1);j>=0;--j)
				{
					if(line[j]->character == ' ')
					{
						//Update i, remove characters following the space from line.
						i -= (line.size()-j-1);
						for(k=(line.size()-1);k>=j;--k)
							line.erase(line.begin()+k);

						foundSpace = true;

						break;
					}
				}
				//If there was no space, just break the line at the last character.
				if(!foundSpace)
				{
					--i;
					line.erase(line.begin()+(line.size()-1));
				}

				//Draw the line.
				tempX = x;
				for(k=0;k<line.size();++k)
				{
					if(line[k])
					{
						drawGlyph(line[k]->xPos,
								  line[k]->yPos,
								  line[k]->width,
								  scale,
								  tempX + line[k]->xOffset,
								  y,
								  r,
								  g,
								  b,
								  a);
						tempX += (line[k]->advance * scale);
					}
					else
						tempX += (glyphs[getGlyphIndex(' ')].advance * scale);
				}

				y += height;
				span = 0.0f;
				line.clear();
			}
		}
	}

	//Draw the last line.
	if(line.size() > 0)
	{
		//Draw the line.
		tempX = x;
		for(k=0;k<line.size();++k)
		{
			if(line[k])
			{
				drawGlyph(line[k]->xPos,
						  line[k]->yPos,
						  line[k]->width,
						  scale,
						  tempX + line[k]->xOffset,
						  y,
						  r,
						  g,
						  b,
						  a);
				tempX += (line[k]->advance * scale);
			}
			else
				tempX += (glyphs[getGlyphIndex(' ')].advance * scale);
		}
	}
#endif
}

//------------------------------------------------------------------------------
void OpenGLFont::drawAtlas()
{
	glBindTexture(GL_TEXTURE_2D, atlas);

	glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
	glBegin(GL_QUADS);
	{
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 0.0f,
							 0.0f);
		glVertex2f(0.0f,
				   0.0f);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 1.0f,
							 0.0f);
		glVertex2f(800.0f,
				   0.0f);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 1.0f,
							 1.0f);
		glVertex2f(800.0f,
				   800.0f);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 0.0f,
							 1.0f);
		glVertex2f(0.0f,
				   800.0f);
	}
	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
}

//------------------------------------------------------------------------------
const FilePath OpenGLFont::getFontPath(const string& filename)
{
	int i;
	FilePath retval;
	std::vector<FilePath> fontDirs;

	fontDirs = FilePath::getFontsDirs();
	fontDirs.push_back(FilePath::getDataFilesDir().getChild("fonts"));

	for(i=0;i<fontDirs.size();++i)
	{
		FilePath tempPath;

		tempPath = fontDirs[i].getChild(filename);
		if(tempPath.isFile())
		{
			retval = tempPath;
			break;
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
void OpenGLFont::copyGlyph(uint8_t *atlasData,
						   SDL_Surface *glyph,
						   int x,
						   int y,
						   int yOffset)
{
	int i, j;
	uint8_t *pix;
	int xStart, yStart;

	pix = (uint8_t *)glyph->pixels;
	xStart = x * 4;
	yStart = (y + yOffset - 1) * (atlasSize*4);
	for(j=0;j<glyph->h;++j)
	{
//#if __BIG_ENDIAN__==1
#ifdef __BIG_ENDIAN__
		for(i=0;i<(glyph->w * 4);i+=4)
		{
			atlasData[yStart+xStart+i] = pix[(j*glyph->w*4)+i+3];
			atlasData[yStart+xStart+i+1] = pix[(j*glyph->w*4)+i+2];
			atlasData[yStart+xStart+i+2] = pix[(j*glyph->w*4)+i+1];
			atlasData[yStart+xStart+i+3] = pix[(j*glyph->w*4)+i];
		}
#else
		for(i=0;i<(glyph->w * 4);++i)
			atlasData[yStart+xStart+i] = pix[(j*glyph->w*4)+i];
#endif
		yStart += (atlasSize*4);
	}
}

//------------------------------------------------------------------------------
void OpenGLFont::drawGlyph(float atlasX,
						   float atlasY,
						   float width,
						   float scale,
						   float x,
						   float y,
						   float r,
						   float g,
						   float b,
						   float a)
{
	glBindTexture(GL_TEXTURE_2D, atlas);

	glColor4f(r, g, b, a);
	glBegin(GL_QUADS);
	{
		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 atlasX/atlasSize,
							 atlasY/atlasSize);
		glVertex2f(x,
				   y);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 (atlasX+width)/atlasSize,
							 atlasY/atlasSize);
		glVertex2f(x+(width*scale),
				   y);

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 (atlasX+width)/atlasSize,
							 (atlasY+height+2)/atlasSize);
		glVertex2f(x+(width*scale),
				   y+((height+2)*scale));

		glMultiTexCoord2fARB(GL_TEXTURE0_ARB,
							 atlasX/atlasSize,
							 (atlasY+height+2)/atlasSize);
		glVertex2f(x,
				   y+((height+2)*scale));
	}
	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
}

//------------------------------------------------------------------------------
uint32_t OpenGLFont::powerOfTwo(uint32_t val)
{
	uint32_t retval = 1;

	//Surely there's a better way to do this?
	while(retval < val)
		retval *= 2;

	return retval;
}

//------------------------------------------------------------------------------
const char OpenGLFont::characters[] = {
									   'a',
									   'b',
									   'c',
									   'd',
									   'e',
									   'f',
									   'g',
									   'h',
									   'i',
									   'j',
									   'k',
									   'l',
									   'm',
									   'n',
									   'o',
									   'p',
									   'q',
									   'r',
									   's',
									   't',
									   'u',
									   'v',
									   'w',
									   'x',
									   'y',
									   'z',
									   'A',
									   'B',
									   'C',
									   'D',
									   'E',
									   'F',
									   'G',
									   'H',
									   'I',
									   'J',
									   'K',
									   'L',
									   'M',
									   'N',
									   'O',
									   'P',
									   'Q',
									   'R',
									   'S',
									   'T',
									   'U',
									   'V',
									   'W',
									   'X',
									   'Y',
									   'Z',
									   '0',
									   '1',
									   '2',
									   '3',
									   '4',
									   '5',
									   '6',
									   '7',
									   '8',
									   '9',
									   '+',
									   '-',
									   '=',
									   '.',
									   ',',
									   ';',
									   ':',
									   '\'',
									   '?',
									   '(',
									   ')',
									   '!',
									   '\"',
									   '\\',
									   '/',
									   '<',
									   '>',
									   '_',
									   ' '
									  };

//------------------------------------------------------------------------------
const size_t OpenGLFont::charactersSize = sizeof(characters);

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
TTFSingleton& TTFSingleton::getInstance()
{
	static TTFSingleton retval;

	return retval;
}

//------------------------------------------------------------------------------
TTFSingleton::TTFSingleton()
{
#ifdef USE_TTF
	if(TTF_Init() == -1)
		cout << "OpenGLFont: Could not initialise SDL_ttf.";
#endif
}

//------------------------------------------------------------------------------
TTFSingleton::~TTFSingleton()
{
#ifdef USE_TTF
	TTF_Quit();
#endif
}
