//	StreamingSound.cpp - Class used to stream multiple sounds off disk.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2010  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "StreamingSound.h"
#include "MiscHelper.h"
#include "FilePath.h"
#include "Random.h"

#include "SDL.h"
#include <cmath>
#include <cassert>
#include <iostream>

using namespace std;

//------------------------------------------------------------------------------
StreamingSound::StreamingSound():
timeToStop(false),
fade(1.0f),
numSoundsLoaded(0),
loadSoundsCount(0)
{
	int i;
	FilePath phil = FilePath::getDataFilesDir().getChild("sounds");

	//Fill out paths.
	for(i=0;i<phil.getNumChildren();++i)
		paths.push_back(phil.getChild(i));

	startThread(this);
}

//------------------------------------------------------------------------------
StreamingSound::~StreamingSound()
{
	timeToStop = true;

	//To make sure the thread's quit.
	SDL_Delay(300);
}

//------------------------------------------------------------------------------
void StreamingSound::update(float hide1, float hide2, float hide3)
{
	int i;
	const int ritual1Sounds = NumSounds/3;
	const int ritual2Sounds = ritual1Sounds*2;

	if(numSoundsLoaded < NumSounds)
	{
		--loadSoundsCount;
		if(loadSoundsCount <= 0)
		{
			triggerSound();
			++numSoundsLoaded;

			loadSoundsCount = 2;
		}
	}
	else
	{
		for(i=0;i<NumSounds;++i)
		{
			if(!sounds[i].getActive())
				triggerSound();
		}
	}

	//Fade out sounds as the rituals clear the noise.
	for(i=0;i<ritual1Sounds;++i)
		sounds[i].setLevel(1.0f-hide1);
	for(i=ritual1Sounds;i<ritual2Sounds;++i)
		sounds[i].setLevel(1.0f-hide2);
	for(i=ritual2Sounds;i<NumSounds;++i)
		sounds[i].setLevel(1.0f-hide3);
}

//------------------------------------------------------------------------------
void StreamingSound::getAudio(float *buffer, int numSamples)
{
	int i;
	const float scale = (1.0f/(float)NumSounds) * 8.0f;

	for(i=0;i<NumSounds;++i)
	{
		if(sounds[i].getActive())
			sounds[i].getAudio(buffer, numSamples);
	}

	//Scale sounds and apply fade if necessary, distort.
	for(i=0;i<numSamples;++i)
	{
		buffer[i*2] *= scale * fakeLog(fade);
		buffer[(i*2)+1] *= scale * fakeLog(fade);

		buffer[i*2] = tanhf(buffer[i*2] * 2.0f) * 0.99f;
		buffer[(i*2)+1] = tanhf(buffer[(i*2)+1] * 2.0f) * 0.99f;
	}
}

//------------------------------------------------------------------------------
bool StreamingSound::run()
{
	int i;

	while(!timeToStop)
	{
		//Load buffers.
		for(i=0;i<NumSounds;++i)
		{
			if(sounds[i].getActive())
				sounds[i].updateBuffers();
		}

		SDL_Delay(62);
	}

	return true;
}

//------------------------------------------------------------------------------
uint32_t StreamingSound::triggerSound()
{
	int i;
	int tempint;

	assert(paths.size() > 0);

	for(i=0;i<NumSounds;++i)
	{
		if(!sounds[i].getActive() && (paths.size() > 0))
		{
			tempint = (int)(Random::getGlobal().getRandomFloat() * paths.size() * 0.999999f);
			sounds[i].load(paths[tempint]);
			break;
		}
	}
	assert(i < NumSounds);

	return i+1;
}

//------------------------------------------------------------------------------
void StreamingSound::stopSound(uint32_t index)
{
	if(sounds[index-1].getActive())
		sounds[index-1].stopPlaying();
}

//------------------------------------------------------------------------------
void StreamingSound::setFade(float val)
{
	fade = val;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
StreamingSound::Sound::Sound():
active(false),
fade(0.0f),
fadeDir(true),
file(0),
filePosition(0),
readBuffer(0),
readPosition(0)
{
	panVal = Random::getGlobal().getRandomFloat() * 6.283185f;
	panSpeed = 0.01f + (Random::getGlobal().getRandomFloat() * 0.01f);
}

//------------------------------------------------------------------------------
StreamingSound::Sound::~Sound()
{
	if(file)
		ov_clear(&vorbisFile);
}

//------------------------------------------------------------------------------
void StreamingSound::Sound::getAudio(float *buffer, int numSamples)
{
	int i;
	float pan;
	int tempint;

	if(!file)
		return;

	panVal += panSpeed;
	if(panVal > 6.283185f)
		panVal -= 6.283185f;
	pan = (sinf(panVal) * 0.5f) + 0.5f;

	//If we've not reached the end of this buffer.
	if((BufferSize-readPosition) >= numSamples)
	{
		for(i=0;i<numSamples;++i)
		{
			buffer[(i*2)] += left[readBuffer][readPosition] * fade * (1.0f-pan) * level;
			buffer[(i*2)+1] += left[readBuffer][readPosition] * fade * pan * level;
			if((fade < 1.0f) && fadeDir)
				fade += 0.00001f;
			else if((fade > 0.0f) && !fadeDir)
				fade -= 0.00001f;

			++readPosition;
		}
		if(readPosition >= length)
		{
			readPosition = 0;
			clearedBuffers[readBuffer] = true;
			readBuffer = (readBuffer+1)%NumBuffers;
		}
	}
	//If we need to read from the next buffer halfway through this audio block.
	else
	{
		tempint = BufferSize-readPosition;

		//Read the rest of the samples from the current buffer.
		for(i=0;i<tempint;++i)
		{
			buffer[(i*2)] += left[readBuffer][readPosition] * fade * (1.0f-pan) * level;
			buffer[(i*2)+1] += left[readBuffer][readPosition] * fade * pan * level;
			if((fade < 1.0f) && fadeDir)
				fade += 0.00001f;
			else if((fade > 0.0f) && !fadeDir)
				fade -= 0.00001f;

			++readPosition;
		}
		//tempint = numSamples - tempint;

		//Move to the next buffer.
		readPosition = 0;
		clearedBuffers[readBuffer] = true;
		readBuffer = (readBuffer+1)%NumBuffers;

		//Read the remaining samples from the new buffer.
		for(i=tempint;i<numSamples;++i)
		{
			buffer[(i*2)] += left[readBuffer][readPosition] * fade * (1.0f-pan) * level;
			buffer[(i*2)+1] += left[readBuffer][readPosition] * fade * pan * level;
			if((fade < 1.0f) && fadeDir)
				fade += 0.00001f;
			else if((fade > 0.0f) && !fadeDir)
				fade -= 0.00001f;

			++readPosition;
		}
	}

	if((fade <= 0.0f) && !fadeDir)
		active = false;
}

//------------------------------------------------------------------------------
void StreamingSound::Sound::setLevel(float val)
{
	level = val;
}

//------------------------------------------------------------------------------
void StreamingSound::Sound::updateBuffers()
{
	int i;

	for(i=0;i<NumBuffers;++i)
	{
		if(clearedBuffers[i])
		{
			fillBuffer(i);
			clearedBuffers[i] = false;
		}
	}
}

//------------------------------------------------------------------------------
void StreamingSound::Sound::load(const FilePath& soundFile)
{
	int i;
	int result = 0;

	if(file)
	{
		ov_clear(&vorbisFile);
		file = 0;
	}

	for(i=0;i<NumBuffers;++i)
		clearedBuffers[i] = true;

	file = fopen(soundFile.getPath().c_str(), "rb");
	if(!file)
		cout << "vorbisfile: fopen could not open file: " << soundFile.getPath() << endl;
	else
	{
		if(ov_open(file, &vorbisFile, NULL, 0) < 0)
			cout << "vorbisfile: Could not open file: " << soundFile.getPath() << endl;
		else
		{
			if(ov_streams(&vorbisFile) > 0)
			{
				vorbis_info *info;

				cout << "vorbisfile: Loaded file: " << soundFile.getPath() << endl;

				info = ov_info(&vorbisFile, 0);
				if(info->channels > 0)
				{
					//Fill the buffers.
					for(i=0;i<NumBuffers;++i)
						fillBuffer(i);

					//Get the length of the file.
					length = ov_pcm_total(&vorbisFile, 0);

					fade = 0.0f;
					fadeDir = true;
					active = true;
				}
				else
				{
					ov_clear(&vorbisFile);
					cout << "vorbisfile: " << soundFile.getPath() << " is mono, so ignoring." << endl;
				}
			}
			else
			{
				ov_clear(&vorbisFile);
				cout << "vorbisfile: There are no streams in this file?" << endl;
			}
		}
	}
}

//------------------------------------------------------------------------------
void StreamingSound::Sound::stopPlaying()
{
	fadeDir = false;
}

//------------------------------------------------------------------------------
void StreamingSound::Sound::fillBuffer(int buf)
{
	int i;
	int result;
	float **tempBuf;
	int currentBitstream = 0;
	int samplesRead = 0;

	//Make sure we're at the correct point to read from.
	ov_pcm_seek(&vorbisFile, filePosition);

	//Read audio from disk into the current block.
	while(samplesRead < BufferSize)
	{
		result = ov_read_float(&vorbisFile,
							   &tempBuf,
							   (BufferSize-samplesRead),
							   &currentBitstream);
		if((currentBitstream > 0)||(result <= 0))
		{
			//We've either gone past the end of our bitstream, or there are
			//no other bitstreams in the file, and we've reached the end of
			//the file.
			filePosition = 0;
			ov_pcm_seek(&vorbisFile, 0);

			//Fade out, disable this Sound.
			fadeDir = false;
		}
		else if(result <= BufferSize)
		{
			//Copy the audio data to the output buffer.
			for(i=0;i<result;++i)
			{
				if(samplesRead >= BufferSize)
					break;
				left[buf][samplesRead] = tempBuf[0][i];
				//right[buf][samplesRead] = tempBuf[0][(i*2)+1];
				//right[buf][samplesRead] = left[buf][samplesRead];
				//right[buf][samplesRead] = tempBuf[1][i];
				++samplesRead;
				++filePosition;
			}
		}
	}
}
