//	StreamingSound.h - Class used to stream multiple sounds off disk.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2010  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef STREAMINGSOUND_H_
#define STREAMINGSOUND_H_

#include "Thread.h"

#include <vector>
#include <stdint.h>
#include <vorbis/vorbisfile.h>

class FilePath;

///	Class used to stream multiple sounds off disk.
class StreamingSound : public Thread
{
  public:
	///	Constructor.
	StreamingSound();
	///	Destructor.
	~StreamingSound();

	///	Called from the main gui thread to periodically load new sounds.
	void update(float hide1, float hide2, float hide3);

	///	Returns the current block of audio.
	void getAudio(float *buffer, int numSamples);

	///	The thread where more audio is loaded from disk.
	bool run();

	///	Triggers a random sound, returns an identifier to identify it with.
	uint32_t triggerSound();
	///	Tells the indexed sound to stop.
	void stopSound(uint32_t index);

	///	Fades the sound when the game starts to quit.
	void setFade(float val);
  private:
	///	Class representing a single sound.
	class Sound
	{
	  public:
		///	Constructor.
		Sound();
		///	Destructor.
		~Sound();

		///	Returns this Sound's audio.
		void getAudio(float *buffer, int numSamples);

		///	Sets this Sound's level.
		void setLevel(float val);

		///	Called from the thread to update the Sound's streaming buffers.
		void updateBuffers();

		///	Returns whether or not this Sound's currently active.
		bool getActive() const {return active;};
		///	Loads the first part of a sound file, makes the Sound active.
		void load(const FilePath& soundFile);
		///	Tells the Sound to stop playing.
		/*!
			It'll fade out, so it won't become inactive immediately.
		 */
		void stopPlaying();
	  private:
		///	Helper method to fill a buffer with audio.
		void fillBuffer(int buf);

		///	Some constants.
		enum
		{
			BufferSize = 11025,
			NumBuffers = 3
		};

		///	Whether or not the Sound is active.
		bool active;
		///	Used to fade the Sound in and out.
		float fade;
		///	Used to fade the Sound in and out.
		bool fadeDir;

		///	This sound's level.
		float level;
		///	Used to pan the Sound left and right.
		float panVal;
		///	Used to pan the Sound left and right.
		float panSpeed;

		///	The sound file to play.
		FILE *file;
		///	The sound file to play.
		OggVorbis_File vorbisFile;
		///	Our read position within the file.
		unsigned int filePosition;
		///	The length of the file in samples.
		unsigned int length;

		///	Temporary buffers used to stream data from the disk.
		float left[NumBuffers][BufferSize];
		///	Temporary buffers used to stream data from the disk.
		float right[NumBuffers][BufferSize];
		///	Which buffer we're currently reading from.
		int readBuffer;
		///	Which buffers need re-loaded.
		bool clearedBuffers[NumBuffers];
		///	Our read position in the active buffer.
		int readPosition;
	};

	///	The total number of Sounds to use.
	enum
	{
		NumSounds = 24
	};
	///	All the Sounds which may be used at once.
	Sound sounds[NumSounds];

	///	The paths of all the sound files we can use.
	std::vector<FilePath> paths;

	///	Tells the streaming thread to quit.
	bool timeToStop;

	///	The fade out value.
	float fade;

	///	Used to load sounds gradually on startup.
	int numSoundsLoaded;
	///	Used to load sounds gradually on startup.
	int loadSoundsCount;
};

#endif
