//	StringModel.cpp - Simple physically-modelled string class.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2011  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "StringModel.h"

#include <cmath>

//------------------------------------------------------------------------------
StringModel::StringModel():
stringOutput(0.0f),
bridgeFilter(0.0f),
decay(0.66f),
input(0),
waitingInput(0.0f)
{
	int i;
	static bool alternateSizes = false;

	if(alternateSizes)
		stringLength = 64;
	else
		stringLength = 128;
	alternateSizes = !alternateSizes;

	inputPos = stringLength/4;
	outputPos = inputPos * 3;

	delayLine[0] = new float[stringLength];
	delayLine[1] = new float[stringLength];

	for(i=0;i<stringLength;++i)
	{
		delayLine[0][i] = 0.0f;
		delayLine[1][i] = 0.0f;
	}
}

//------------------------------------------------------------------------------
StringModel::~StringModel()
{
	delete delayLine[0];
	delete delayLine[1];
}

//------------------------------------------------------------------------------
float StringModel::getSample()
{
	int i;
	float tempf;

	//Apply other StringModel input.
	{
		ScopedLock lock(critSec);

		if(input)
		{
			delayLine[0][inputPos] += input->stringOutput * 0.8f;
			delayLine[1][inputPos] -= input->stringOutput * 0.8f;
		}
	}

	//Apply arbitrary input.
	delayLine[0][inputPos] += waitingInput;
	delayLine[1][inputPos] -= waitingInput;
	waitingInput = 0.0f;

	//Update the string model.
	for(i=(stringLength-1);i>0;--i)
	{
		delayLine[0][i] = delayLine[0][i-1];
		delayLine[1][i] = delayLine[1][i-1];
	}

	//Nut wrap-around.
	delayLine[0][0] = -delayLine[1][stringLength-1];

	//Bridge wrap-around.
	tempf = delayLine[0][stringLength-1];
	tempf = bridgeFilter + (1.0f-decay)*(tempf - bridgeFilter);
	delayLine[1][0] = -tempf;
	bridgeFilter = tempf;

	stringOutput = tanhf(delayLine[0][outputPos] + delayLine[1][outputPos]);

	return stringOutput;
}

//------------------------------------------------------------------------------
void StringModel::setInputString(StringModel *other)
{
	input = other;
}

//------------------------------------------------------------------------------
void StringModel::applyInput(float val)
{
	ScopedLock lock(critSec);

	waitingInput = val;
}
