//	SimpleSynth.h - Simple analogue-esque synth.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2010  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef SIMPLESYNTH_H_
#define SIMPLESYNTH_H_

#include "MonophonicSynth.h"
#include "MoogFilter.h"

#include <cmath>

///	Simple analogue-esque synth.
class SimpleSynth : public MonophonicSynth
{
  public:
	///	Constructor.
	SimpleSynth();
	///	Destructor.
	~SimpleSynth();

	///	Generates the synth's audio.
	void getAudio(float *buffer, int numSamples);

	///	Triggers a note.
	void triggerNote(float freq);
	///	Sets a parameter.
	void setParameter(const std::string& name, float value, bool direct = false);
	///	Returns the number of parameters this synth has.
	int getNumParameters() const {return 15;};
	///	Returns the value of the indexed parameter.
	float getParameter(int i) const;
	///	Returns the name of the indexed parameter.
	std::string getParameterName(int i) const;
  private:
	///	Singleton used to create the saw and square wavetables.
	class WavetableSingleton
	{
	  public:
		///	Returns the sole instance of the singleton.
		static WavetableSingleton& getInstance();

		///	Returns a pointer to the sawtooth wavetable.
		const float * const getSawTable() const {return sawTable;};
		///	Returns a pointer to the square wavetable.
		const float * const getSquareTable() const {return squareTable;};
		///	Returns the size of the wavetables.
		int getTableSize() const {return TableSize;};
#ifdef WIN32
		///	Destructor.
		~WavetableSingleton() {};
#endif
	  private:
		///	The size of the wavetables.
		enum
		{
			TableSize = 1024
		};

		///	Constructor.
		WavetableSingleton()
		{
			int i, j;
			float index;

			//Generate the sawtooth wavetable.
			j = 1.0f;
			while(j < 128.0f)
			{
				index = 0.0f;
				for(i=0;i<TableSize;++i)
				{
					sawTable[i] += sinf(index * j * 6.283185f) * (1.0f/j) * 0.5f;
					index += (1.0f/(float)TableSize);
				}
				j += 1.0f;
			}

			//Generate the square wave wavetable.
			j = 1.0f;
			while(j < 256.0f)
			{
				index = 0.0f;
				for(i=0;i<TableSize;++i)
				{
					squareTable[i] += sinf(index * j * 6.283185f) * (1.0f/j) * 0.5f;
					index += (1.0f/(float)TableSize);
				}
				j += 2.0f;
			}
		};
#ifndef WIN32
		///	Destructor.
		~WavetableSingleton() {};
#endif
		///	The sawtooth wavetable.
		float sawTable[TableSize];
		///	The square wavetable.
		float squareTable[TableSize];
	};

	//--Manipulatable Parameters--
	///	How much of the 2 waveforms is mixed in.
	float waveAmount;
	///	The filter's cutoff frequency.
	float filterCutoff;
	///	The filter's resonance.
	float filterRes;
	///	Whether the filter attack's an exp or a log curve.
	bool filterAttackCurve;
	///	Whether the filter release's an exp or a log curve.
	bool filterReleaseCurve;
	///	The filter's attack time.
	float filterAttackTime;
	///	The filter's release time.
	float filterReleaseTime;
	///	The filter's envelope amount (-1 -> +1).
	float filterEnvAmount;
	///	Whether the attack's an exp or a log curve.
	bool attackCurve;
	///	Whether the release's an exp or a log curve.
	bool releaseCurve;
	///	The attack time.
	float attackTime;
	///	The release time.
	float releaseTime;
	///	The time it takes to play a note.
	float noteLength;
	///	The amount of gain to apply before the distortion.
	float distGain;
	///	The octave to play notes at.
	float octave;

	//--Internal Parameters--
	///	The size of the wavetables.
	int tableSize;
	///	The sawtooth wavetable.
	const float *sawTable;
	///	The square wavetable.
	const float *squareTable;
	///	Our current position within the wavetables.
	float tableIndex;

	///	The current frequency.
	float frequency;

	///	The lowpass filter.
	MoogFilter filter;

	///	Our current position within noteLength.
	float noteIndex;
	///	The length (in seconds) of a single sample.
	float increment;
};

#endif
